@@ -1,4 +1,4 @@
-/*	$OpenBSD: cert.c,v 1.47 2021/11/05 10:50:41 claudio Exp $ */
+/*	$OpenBSD: cert.c,v 1.56 2022/02/04 16:50:49 tb Exp $ */
 /*
  * Copyright (c) 2021 Job Snijders <job@openbsd.org>
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
@@ -29,6 +29,7 @@
 
 #include <openssl/asn1.h>
 #include <openssl/x509.h>
+#include <openssl/x509v3.h>
 
 #include "extern.h"
 
@@ -47,31 +48,21 @@ struct	parse {
 	const char	*fn; /* currently-parsed file */
 };
 
-static ASN1_OBJECT	*carepo_oid;	/* 1.3.6.1.5.5.7.48.5 (caRepository) */
-static ASN1_OBJECT	*mft_oid;	/* 1.3.6.1.5.5.7.48.10 (rpkiManifest) */
-static ASN1_OBJECT	*notify_oid;	/* 1.3.6.1.5.5.7.48.13 (rpkiNotify) */
-
-static void
-cert_init_oid(void)
-{
-	if ((carepo_oid = OBJ_txt2obj("1.3.6.1.5.5.7.48.5", 1)) == NULL)
-		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.48.5");
-	if ((mft_oid = OBJ_txt2obj("1.3.6.1.5.5.7.48.10", 1)) == NULL)
-		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.48.10");
-	if ((notify_oid = OBJ_txt2obj("1.3.6.1.5.5.7.48.13", 1)) == NULL)
-		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.48.13");
-}
+extern ASN1_OBJECT	*certpol_oid;	/* id-cp-ipAddr-asNumber cert policy */
+extern ASN1_OBJECT	*carepo_oid;	/* 1.3.6.1.5.5.7.48.5 (caRepository) */
+extern ASN1_OBJECT	*manifest_oid;	/* 1.3.6.1.5.5.7.48.10 (rpkiManifest) */
+extern ASN1_OBJECT	*notify_oid;	/* 1.3.6.1.5.5.7.48.13 (rpkiNotify) */
 
 /*
  * Append an IP address structure to our list of results.
- * This will also constrain us to having at most one inheritence
- * statement per AFI and also not have overlapping rages (as prohibited
+ * This will also constrain us to having at most one inheritance
+ * statement per AFI and also not have overlapping ranges (as prohibited
  * in section 2.2.3.6).
  * It does not make sure that ranges can't coalesce, that is, that any
  * two ranges abut each other.
  * This is warned against in section 2.2.3.6, but doesn't change the
  * semantics of the system.
- * Return zero on failure (IP overlap) non-zero on success.
+ * Returns zero on failure (IP overlap) non-zero on success.
  */
 static int
 append_ip(struct parse *p, const struct cert_ip *ip)
@@ -113,7 +104,7 @@ append_as(struct parse *p, const struct cert_as *as)
 
 /*
  * Construct a RFC 3779 2.2.3.8 range by its bit string.
- * Return zero on failure, non-zero on success.
+ * Returns zero on failure, non-zero on success.
  */
 static int
 sbgp_addr(struct parse *p,
@@ -270,12 +261,9 @@ sbgp_sia_resource_entry(struct parse *p,
 	if (!ASN1_frame(p->fn, dsz, &d, &plen, &ptag))
 		goto out;
 
-	if (carepo_oid == NULL)
-		cert_init_oid();
-
 	if (OBJ_cmp(oid, carepo_oid) == 0)
 		rc = sbgp_sia_resource_carepo(p, d, plen);
-	else if (OBJ_cmp(oid, mft_oid) == 0)
+	else if (OBJ_cmp(oid, manifest_oid) == 0)
 		rc = sbgp_sia_resource_mft(p, d, plen);
 	else if (OBJ_cmp(oid, notify_oid) == 0)
 		rc = sbgp_sia_resource_notify(p, d, plen);
@@ -588,6 +576,12 @@ sbgp_assysnum(struct parse *p, X509_EXTENSION *ext)
 	int			 dsz, rc = 0, i, ptag;
 	long			 plen;
 
+	if (!X509_EXTENSION_get_critical(ext)) {
+		cryptowarnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
+		    "extension not critical", p->fn);
+		goto out;
+	}
+
 	if ((dsz = i2d_X509_EXTENSION(ext, &sv)) < 0) {
 		cryptowarnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
 		    "failed extension parse", p->fn);
@@ -687,7 +681,7 @@ out:
 
 /*
  * Parse RFC 3779 2.2.3.9 range of addresses.
- * Return zero on failure, non-zero on success.
+ * Returns zero on failure, non-zero on success.
  */
 static int
 sbgp_addr_range(struct parse *p, struct cert_ip *ip,
@@ -891,6 +885,12 @@ sbgp_ipaddrblk(struct parse *p, X509_EXTENSION *ext)
 	const ASN1_TYPE		*t = NULL;
 	int			 i;
 
+	if (!X509_EXTENSION_get_critical(ext)) {
+		cryptowarnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
+		    "extension not critical", p->fn);
+		goto out;
+	}
+
 	if ((dsz = i2d_X509_EXTENSION(ext, &sv)) < 0) {
 		cryptowarnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
 		    "failed extension parse", p->fn);
@@ -971,6 +971,84 @@ out:
 	return rc;
 }
 
+/*
+ * Parse the certificate policies extension and check that it follows RFC 7318.
+ * Returns zero on failure, non-zero on success.
+ */
+static int
+certificate_policies(struct parse *p, X509_EXTENSION *ext)
+{
+	STACK_OF(POLICYINFO)		*policies = NULL;
+	POLICYINFO			*policy;
+	STACK_OF(POLICYQUALINFO)	*qualifiers;
+	POLICYQUALINFO			*qualifier;
+	int				 nid;
+	int				 rc = 0;
+
+	if (!X509_EXTENSION_get_critical(ext)) {
+		cryptowarnx("%s: RFC 6487 section 4.8.9: certificatePolicies: "
+		    "extension not critical", p->fn);
+		goto out;
+	}
+
+	if ((policies = X509V3_EXT_d2i(ext)) == NULL) {
+		cryptowarnx("%s: RFC 6487 section 4.8.9: certificatePolicies: "
+		    "failed extension parse", p->fn);
+		goto out;
+	}
+
+	if (sk_POLICYINFO_num(policies) != 1) {
+		warnx("%s: RFC 6487 section 4.8.9: certificatePolicies: "
+		    "want 1 policy, got %d", p->fn,
+		    sk_POLICYINFO_num(policies));
+		goto out;
+	}
+
+	policy = sk_POLICYINFO_value(policies, 0);
+	assert(policy != NULL && policy->policyid != NULL);
+
+	if (OBJ_cmp(policy->policyid, certpol_oid) != 0) {
+		char pbuf[128], cbuf[128];
+
+		OBJ_obj2txt(pbuf, sizeof(pbuf), policy->policyid, 1);
+		OBJ_obj2txt(cbuf, sizeof(cbuf), certpol_oid, 1);
+		warnx("%s: RFC 7318 section 2: certificatePolicies: "
+		    "unexpected OID: %s, want %s", p->fn, pbuf, cbuf);
+		goto out;
+	}
+
+	/* Policy qualifiers are optional. If they're absent, we're done. */
+	if ((qualifiers = policy->qualifiers) == NULL) {
+		rc = 1;
+		goto out;
+	}
+
+	if (sk_POLICYQUALINFO_num(qualifiers) != 1) {
+		warnx("%s: RFC 7318 section 2: certificatePolicies: "
+		    "want 1 policy qualifier, got %d", p->fn,
+		    sk_POLICYQUALINFO_num(qualifiers));
+		goto out;
+	}
+
+	qualifier = sk_POLICYQUALINFO_value(qualifiers, 0);
+	assert(qualifier != NULL && qualifier->pqualid != NULL);
+
+	if ((nid = OBJ_obj2nid(qualifier->pqualid)) != NID_id_qt_cps) {
+		warnx("%s: RFC 7318 section 2: certificatePolicies: "
+		    "want CPS, got %d (%s)", p->fn, nid, OBJ_nid2sn(nid));
+		goto out;
+	}
+
+	if (verbose > 1)
+		warnx("%s: CPS %.*s", p->fn, qualifier->d.cpsuri->length,
+		    qualifier->d.cpsuri->data);
+
+	rc = 1;
+ out:
+	sk_POLICYINFO_pop_free(policies, POLICYINFO_free);
+	return rc;
+}
+
 /*
  * Parse and partially validate an RPKI X509 certificate (either a trust
  * anchor or a certificate) as defined in RFC 6487.
@@ -980,8 +1058,7 @@ out:
  * is also dereferenced.
  */
 static struct cert *
-cert_parse_inner(X509 **xp, const char *fn, const unsigned char *der,
-    size_t len, int ta)
+cert_parse_inner(const char *fn, const unsigned char *der, size_t len, int ta)
 {
 	int		 rc = 0, extsz, c;
 	int		 sia_present = 0;
@@ -991,8 +1068,6 @@ cert_parse_inner(X509 **xp, const char *fn, const unsigned char *der,
 	ASN1_OBJECT	*obj;
 	struct parse	 p;
 
-	*xp = NULL;
-
 	/* just fail for empty buffers, the warning was printed elsewhere */
 	if (der == NULL)
 		return NULL;
@@ -1002,7 +1077,7 @@ cert_parse_inner(X509 **xp, const char *fn, const unsigned char *der,
 	if ((p.res = calloc(1, sizeof(struct cert))) == NULL)
 		err(1, NULL);
 
-	if ((x = *xp = d2i_X509(NULL, &der, len)) == NULL) {
+	if ((x = d2i_X509(NULL, &der, len)) == NULL) {
 		cryptowarnx("%s: d2i_X509_bio", p.fn);
 		goto out;
 	}
@@ -1030,6 +1105,9 @@ cert_parse_inner(X509 **xp, const char *fn, const unsigned char *der,
 			sia_present = 1;
 			c = sbgp_sia(&p, ext);
 			break;
+		case NID_certificate_policies:
+			c = certificate_policies(&p, ext);
+			break;
 		case NID_crl_distribution_points:
 			/* ignored here, handled later */
 			break;
@@ -1141,9 +1219,6 @@ cert_parse_inner(X509 **xp, const char *fn, const unsigned char *der,
 		goto out;
 	}
 
-	if (X509_up_ref(x) == 0)
-		errx(1, "%s: X509_up_ref failed", __func__);
-
 	p.res->x509 = x;
 
 	rc = 1;
@@ -1151,51 +1226,67 @@ out:
 	if (rc == 0) {
 		cert_free(p.res);
 		X509_free(x);
-		*xp = NULL;
 	}
 	return (rc == 0) ? NULL : p.res;
 }
 
 struct cert *
-cert_parse(X509 **xp, const char *fn, const unsigned char *der, size_t len)
+cert_parse(const char *fn, const unsigned char *der, size_t len)
 {
-	return cert_parse_inner(xp, fn, der, len, 0);
+	return cert_parse_inner(fn, der, len, 0);
 }
 
 struct cert *
-ta_parse(X509 **xp, const char *fn, const unsigned char *der, size_t len,
+ta_parse(const char *fn, const unsigned char *der, size_t len,
     const unsigned char *pkey, size_t pkeysz)
 {
+	ASN1_TIME	*notBefore, *notAfter;
 	EVP_PKEY	*pk = NULL, *opk = NULL;
 	struct cert	*p;
 	int		 rc = 0;
 
-	if ((p = cert_parse_inner(xp, fn, der, len, 1)) == NULL)
+	if ((p = cert_parse_inner(fn, der, len, 1)) == NULL)
 		return NULL;
 
-	if (pkey != NULL) {
-		assert(*xp != NULL);
-		pk = d2i_PUBKEY(NULL, &pkey, pkeysz);
-		assert(pk != NULL);
-
-		if ((opk = X509_get_pubkey(*xp)) == NULL)
-			cryptowarnx("%s: RFC 6487 (trust anchor): "
-			    "missing pubkey", fn);
-		else if (EVP_PKEY_cmp(pk, opk) != 1)
-			cryptowarnx("%s: RFC 6487 (trust anchor): "
-			    "pubkey does not match TAL pubkey", fn);
-		else
-			rc = 1;
-
-		EVP_PKEY_free(pk);
-		EVP_PKEY_free(opk);
+	/* first check pubkey against the one from the TAL */
+	pk = d2i_PUBKEY(NULL, &pkey, pkeysz);
+	if (pk == NULL) {
+		cryptowarnx("%s: RFC 6487 (trust anchor): bad TAL pubkey", fn);
+		goto badcert;
+	}
+	if ((opk = X509_get0_pubkey(p->x509)) == NULL) {
+		cryptowarnx("%s: RFC 6487 (trust anchor): missing pubkey", fn);
+		goto badcert;
+	} else if (EVP_PKEY_cmp(pk, opk) != 1) {
+		cryptowarnx("%s: RFC 6487 (trust anchor): "
+		    "pubkey does not match TAL pubkey", fn);
+		goto badcert;
 	}
 
+	if ((notBefore = X509_get_notBefore(p->x509)) == NULL) {
+		warnx("%s: certificate has invalid notBefore", fn);
+		goto badcert;
+	}
+	if ((notAfter = X509_get_notAfter(p->x509)) == NULL) {
+		warnx("%s: certificate has invalid notAfter", fn);
+		goto badcert;
+	}
+	if (X509_cmp_current_time(notBefore) != -1) {
+		warnx("%s: certificate not yet valid", fn);
+		goto badcert;
+	}
+	if (X509_cmp_current_time(notAfter) != 1)  {
+		warnx("%s: certificate has expired", fn);
+		goto badcert;
+	}
+
+	rc = 1;
+
+badcert:
+	EVP_PKEY_free(pk);
 	if (rc == 0) {
 		cert_free(p);
 		p = NULL;
-		X509_free(*xp);
-		*xp = NULL;
 	}
 
 	return p;
@@ -1307,7 +1398,7 @@ auth_find(struct auth_tree *auths, const char *aki)
 	return RB_FIND(auth_tree, auths, &a);
 }
 
-int
+void
 auth_insert(struct auth_tree *auths, struct cert *cert, struct auth *parent)
 {
 	struct auth *na;
@@ -1321,8 +1412,6 @@ auth_insert(struct auth_tree *auths, struct cert *cert, struct auth *parent)
 
 	if (RB_INSERT(auth_tree, auths, na) != NULL)
 		err(1, "auth tree corrupted");
-
-	return 1;
 }
 
 static inline int
