@@ -1,4 +1,4 @@
-/*	$OpenBSD: extern.h,v 1.95 2021/11/09 11:03:39 claudio Exp $ */
+/*	$OpenBSD: extern.h,v 1.116 2022/01/28 15:30:23 claudio Exp $ */
 /*
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -148,11 +148,35 @@ struct tal {
 	int		 id; /* ID of this TAL */
 };
 
+/*
+ * Resource types specified by the RPKI profiles.
+ * There might be others we don't consider.
+ */
+enum rtype {
+	RTYPE_INVALID,
+	RTYPE_TAL,
+	RTYPE_MFT,
+	RTYPE_ROA,
+	RTYPE_CER,
+	RTYPE_CRL,
+	RTYPE_GBR,
+	RTYPE_REPO,
+	RTYPE_FILE,
+};
+
+enum location {
+	DIR_UNKNOWN,
+	DIR_TEMP,
+	DIR_VALID,
+};
+
 /*
  * Files specified in an MFT have their bodies hashed with SHA256.
  */
 struct mftfile {
 	char		*file; /* filename (CER/ROA/CRL, no path) */
+	enum rtype	 type; /* file type as determined by extension */
+	enum location	 location;	/* temporary or valid directory */
 	unsigned char	 hash[SHA256_DIGEST_LENGTH]; /* sha256 of body */
 };
 
@@ -162,14 +186,17 @@ struct mftfile {
  * manifest file.
  */
 struct mft {
-	char		*file; /* full path of MFT file */
+	char		*path; /* relative path to directory of the MFT */
 	struct mftfile	*files; /* file and hash */
-	size_t		 filesz; /* number of filenames */
-	int		 stale; /* if a stale manifest */
 	char		*seqnum; /* manifestNumber */
 	char		*aia; /* AIA */
 	char		*aki; /* AKI */
 	char		*ski; /* SKI */
+	time_t		 valid_from;
+	time_t		 valid_until;
+	size_t		 filesz; /* number of filenames */
+	unsigned int	 repoid;
+	int		 stale; /* if a stale manifest */
 };
 
 /*
@@ -179,10 +206,10 @@ struct mft {
  */
 struct roa_ip {
 	enum afi	 afi; /* AFI value */
-	size_t		 maxlength; /* max length or zero */
+	struct ip_addr	 addr; /* the address prefix itself */
 	unsigned char	 min[16]; /* full range minimum */
 	unsigned char	 max[16]; /* full range maximum */
-	struct ip_addr	 addr; /* the address prefix itself */
+	unsigned char	 maxlength; /* max length or zero */
 };
 
 /*
@@ -278,21 +305,7 @@ RB_HEAD(auth_tree, auth);
 RB_PROTOTYPE(auth_tree, auth, entry, authcmp);
 
 struct auth	*auth_find(struct auth_tree *, const char *);
-int		 auth_insert(struct auth_tree *, struct cert *, struct auth *);
-
-/*
- * Resource types specified by the RPKI profiles.
- * There might be others we don't consider.
- */
-enum rtype {
-	RTYPE_EOF = 0,
-	RTYPE_TAL,
-	RTYPE_MFT,
-	RTYPE_ROA,
-	RTYPE_CER,
-	RTYPE_CRL,
-	RTYPE_GBR,
-};
+void		 auth_insert(struct auth_tree *, struct cert *, struct auth *);
 
 enum http_result {
 	HTTP_FAILED,	/* anything else */
@@ -307,10 +320,11 @@ enum rrdp_msg {
 	RRDP_START,
 	RRDP_SESSION,
 	RRDP_FILE,
+	RRDP_CLEAR,
 	RRDP_END,
 	RRDP_HTTP_REQ,
 	RRDP_HTTP_INI,
-	RRDP_HTTP_FIN
+	RRDP_HTTP_FIN,
 };
 
 /*
@@ -336,13 +350,15 @@ enum publish_type {
  * and parsed.
  */
 struct entity {
-	enum rtype	 type;		/* type of entity (not RTYPE_EOF) */
-	char		*file;		/* local path to file */
-	int		 has_data;	/* whether data blob is specified */
+	TAILQ_ENTRY(entity) entries;
+	char		*path;		/* path relative to repository */
+	char		*file;		/* filename or valid repo path */
 	unsigned char	*data;		/* optional data blob */
 	size_t		 datasz; 	/* length of optional data blob */
+	unsigned int	 repoid;	/* repository identifier */
 	int		 talid;		/* tal identifier */
-	TAILQ_ENTRY(entity) entries;
+	enum rtype	 type;		/* type of entity (not RTYPE_EOF) */
+	enum location	 location;	/* which directroy the file lives in */
 };
 TAILQ_HEAD(entityq, entity);
 
@@ -376,6 +392,7 @@ struct stats {
 	size_t	 vrps; /* total number of vrps */
 	size_t	 uniqs; /* number of unique vrps */
 	size_t	 del_files; /* number of files removed in cleanup */
+	size_t	 extra_files; /* number of superfluous files */
 	size_t	 del_dirs; /* number of directories removed in cleanup */
 	size_t	 brks; /* number of BGPsec Router Key (BRK) certificates */
 	struct timeval	elapsed_time;
@@ -388,6 +405,7 @@ struct msgbuf;
 
 /* global variables */
 extern int verbose;
+extern int filemode;
 extern const char *tals[];
 extern const char *taldescs[];
 extern unsigned int talrepocnt[];
@@ -402,19 +420,19 @@ struct tal	*tal_read(struct ibuf *);
 
 void		 cert_buffer(struct ibuf *, const struct cert *);
 void		 cert_free(struct cert *);
-struct cert	*cert_parse(X509 **, const char *, const unsigned char *,
-		    size_t);
-struct cert	*ta_parse(X509 **, const char *, const unsigned char *, size_t,
+struct cert	*cert_parse(const char *, const unsigned char *, size_t);
+struct cert	*ta_parse(const char *, const unsigned char *, size_t,
 		    const unsigned char *, size_t);
 struct cert	*cert_read(struct ibuf *);
 void		 cert_insert_brks(struct brk_tree *, struct cert *);
 
+enum rtype	 rtype_from_file_extension(const char *);
 void		 mft_buffer(struct ibuf *, const struct mft *);
 void		 mft_free(struct mft *);
 struct mft	*mft_parse(X509 **, const char *, const unsigned char *,
 		    size_t);
-int		 mft_check(const char *, struct mft *);
 struct mft	*mft_read(struct ibuf *);
+int		 mft_compare(const struct mft *, const struct mft *);
 
 void		 roa_buffer(struct ibuf *, const struct roa *);
 void		 roa_free(struct roa *);
@@ -438,11 +456,9 @@ struct auth	*valid_ski_aki(const char *, struct auth_tree *,
 		    const char *, const char *);
 int		 valid_ta(const char *, struct auth_tree *,
 		    const struct cert *);
-int		 valid_cert(const char *, struct auth_tree *,
-		    const struct cert *);
-int		 valid_roa(const char *, struct auth_tree *, struct roa *);
-int		 valid_filename(const char *);
-int		 valid_filehash(const char *, const char *, size_t);
+int		 valid_cert(const char *, struct auth *, const struct cert *);
+int		 valid_roa(const char *, struct auth *, struct roa *);
+int		 valid_filehash(int, const char *, size_t);
 int		 valid_uri(const char *, size_t, const char *);
 int		 valid_origin(const char *, const char *);
 
@@ -498,28 +514,31 @@ void		 proc_rrdp(int);
 
 /* Repository handling */
 int		 filepath_add(struct filepath_tree *, char *);
-void		 rrdp_save_state(size_t, struct rrdp_session *);
-int		 rrdp_handle_file(size_t, enum publish_type, char *,
+void		 rrdp_clear(unsigned int);
+void		 rrdp_save_state(unsigned int, struct rrdp_session *);
+int		 rrdp_handle_file(unsigned int, enum publish_type, char *,
 		    char *, size_t, char *, size_t);
-char		*repo_filename(const struct repo *, const char *);
+char		*repo_basedir(const struct repo *, int);
+unsigned int	 repo_id(const struct repo *);
+const char	*repo_uri(const struct repo *);
 struct repo	*ta_lookup(int, struct tal *);
 struct repo	*repo_lookup(int, const char *, const char *);
+struct repo	*repo_byid(unsigned int);
 int		 repo_queued(struct repo *, struct entity *);
 void		 repo_cleanup(struct filepath_tree *);
 void		 repo_free(void);
 
-void		 rsync_finish(size_t, int);
-void		 http_finish(size_t, enum http_result, const char *);
-void		 rrdp_finish(size_t, int);
+void		 rsync_finish(unsigned int, int);
+void		 http_finish(unsigned int, enum http_result, const char *);
+void		 rrdp_finish(unsigned int, int);
 
-void		 rsync_fetch(size_t, const char *, const char *);
-void		 http_fetch(size_t, const char *, const char *, int);
-void		 rrdp_fetch(size_t, const char *, const char *,
+void		 rsync_fetch(unsigned int, const char *, const char *,
+		    const char *);
+void		 http_fetch(unsigned int, const char *, const char *, int);
+void		 rrdp_fetch(unsigned int, const char *, const char *,
 		    struct rrdp_session *);
-void		 rrdp_http_done(size_t, enum http_result, const char *);
-
-int		 repo_next_timeout(int);
-void		 repo_check_timeout(void);
+void		 rrdp_http_done(unsigned int, enum http_result, const char *);
+int		 repo_check_timeout(int);
 
 /* Logging (though really used for OpenSSL errors). */
 
@@ -538,6 +557,7 @@ int		 base64_decode(const unsigned char *, size_t,
 int		 base64_encode_len(size_t, size_t *);
 int		 base64_encode(const unsigned char *, size_t, char **);
 char		*hex_encode(const unsigned char *, size_t);
+int		 hex_decode(const char *, char *, size_t);
 
 
 /* Functions for moving data between processes. */
@@ -555,6 +575,7 @@ struct ibuf	*io_buf_recvfd(int, struct ibuf **);
 
 /* X509 helpers. */
 
+void		 x509_init_oid(void);
 char		*x509_get_aia(X509 *, const char *);
 char		*x509_get_aki(X509 *, int, const char *);
 char		*x509_get_ski(X509 *, const char *);
@@ -629,7 +650,7 @@ int	mkpath(const char *);
 #define MAX_URI_LENGTH		2048
 
 /* Maximum acceptable file size */
-#define MAX_FILE_SIZE		2000000
+#define MAX_FILE_SIZE		4000000
 
 /* Maximum number of FileAndHash entries per manifest. */
 #define MAX_MANIFEST_ENTRIES	100000
@@ -643,7 +664,4 @@ int	mkpath(const char *);
 /* Maximum allowd repositories per tal */
 #define MAX_REPO_PER_TAL	1000
 
-/* Timeout for repository synchronisation, in seconds */
-#define MAX_REPO_TIMEOUT	(15 * 60)
-
 #endif /* ! EXTERN_H */
