@@ -1,4 +1,4 @@
-/*	$OpenBSD: gbr.c,v 1.11 2021/10/26 10:52:50 claudio Exp $ */
+/*	$OpenBSD: gbr.c,v 1.14 2022/01/18 16:24:55 claudio Exp $ */
 /*
  * Copyright (c) 2020 Claudio Jeker <claudio@openbsd.org>
  *
@@ -36,7 +36,7 @@ struct	parse {
 	struct gbr	 *res; /* results */
 };
 
-static ASN1_OBJECT	*gbr_oid;
+extern ASN1_OBJECT	*gbr_oid;
 
 /*
  * Parse a full RFC 6493 file and signed by the certificate "cacert"
@@ -53,14 +53,6 @@ gbr_parse(X509 **x509, const char *fn, const unsigned char *der, size_t len)
 	memset(&p, 0, sizeof(struct parse));
 	p.fn = fn;
 
-	/* OID from section 9.1, RFC 6493. */
-	if (gbr_oid == NULL) {
-		gbr_oid = OBJ_txt2obj("1.2.840.113549.1.9.16.1.35", 1);
-		if (gbr_oid == NULL)
-			errx(1, "OBJ_txt2obj for %s failed",
-			    "1.2.840.113549.1.9.16.1.35");
-	}
-
 	cms = cms_parse_validate(x509, fn, der, len, gbr_oid, &cmsz);
 	if (cms == NULL)
 		return NULL;
