@@ -1,4 +1,4 @@
-/*	$OpenBSD: http.c,v 1.49 2021/11/09 11:00:43 claudio Exp $  */
+/*	$OpenBSD: http.c,v 1.52 2022/01/23 12:09:24 claudio Exp $  */
 /*
  * Copyright (c) 2020 Nils Fisher <nils_fisher@hotmail.com>
  * Copyright (c) 2020 Claudio Jeker <claudio@openbsd.org>
@@ -140,7 +140,7 @@ struct http_request {
 	char			*host;
 	char			*port;
 	const char		*path;	/* points into uri */
-	size_t			 id;
+	unsigned int		 id;
 	int			 outfd;
 	int			 redirect_loop;
 };
@@ -150,7 +150,7 @@ TAILQ_HEAD(http_req_queue, http_request);
 static struct http_conn_list	active = LIST_HEAD_INITIALIZER(active);
 static struct http_conn_list	idle = LIST_HEAD_INITIALIZER(idle);
 static struct http_req_queue	queue = TAILQ_HEAD_INITIALIZER(queue);
-static size_t http_conn_count;
+static unsigned int http_conn_count;
 
 static struct msgbuf msgq;
 static struct sockaddr_storage http_bindaddr;
@@ -161,10 +161,10 @@ static size_t tls_ca_size;
 #endif
 
 /* HTTP request API */
-static void	http_req_new(size_t, char *, char *, int, int);
+static void	http_req_new(unsigned int, char *, char *, int, int);
 static void	http_req_free(struct http_request *);
-static void	http_req_done(size_t, enum http_result, const char *);
-static void	http_req_fail(size_t);
+static void	http_req_done(unsigned int, enum http_result, const char *);
+static void	http_req_fail(unsigned int);
 static int	http_req_schedule(struct http_request *);
 
 /* HTTP connection API */
@@ -213,14 +213,17 @@ http_info(const char *uri)
 }
 
 /*
- * Determine whether the character needs encoding, per RFC1738:
- *	- No corresponding graphic US-ASCII.
- *	- Unsafe characters.
+ * Determine whether the character needs encoding, per RFC2396.
  */
 static int
-unsafe_char(const char *c0)
+to_encode(const char *c0)
 {
-	const char *unsafe_chars = " <>\"#{}|\\^~[]`";
+	/* 2.4.3. Excluded US-ASCII Characters */
+	const char *excluded_chars =
+	    " "         /* space */
+	    "<>#\""     /* delims (modulo "%", see below) */
+	    "{}|\\^[]`" /* unwise */
+	    ;
 	const unsigned char *c = (const unsigned char *)c0;
 
 	/*
@@ -230,16 +233,15 @@ unsafe_char(const char *c0)
 	return (iscntrl(*c) || !isascii(*c) ||
 
 	    /*
-	     * Unsafe characters.
-	     * '%' is also unsafe, if is not followed by two
+	     * '%' is also reserved, if is not followed by two
 	     * hexadecimal digits.
 	     */
-	    strchr(unsafe_chars, *c) != NULL ||
+	    strchr(excluded_chars, *c) != NULL ||
 	    (*c == '%' && (!isxdigit(c[1]) || !isxdigit(c[2]))));
 }
 
 /*
- * Encode given URL, per RFC1738.
+ * Encode given URL, per RFC2396.
  * Allocate and return string to the caller.
  */
 static char *
@@ -256,7 +258,7 @@ url_encode(const char *path)
 	 * final URL.
 	 */
 	for (i = 0; i < length; i++)
-		if (unsafe_char(path + i))
+		if (to_encode(path + i))
 			new_length += 2;
 
 	epath = epathp = malloc(new_length + 1);	/* One more for '\0'. */
@@ -268,7 +270,7 @@ url_encode(const char *path)
 	 * Encode, and copy final URL.
 	 */
 	for (i = 0; i < length; i++)
-		if (unsafe_char(path + i)) {
+		if (to_encode(path + i)) {
 			snprintf(epathp, 4, "%%" "%02x",
 			    (unsigned char)path[i]);
 			epathp += 3;
@@ -509,7 +511,8 @@ http_resolv(struct addrinfo **res, const char *host, const char *port)
  * Create and queue a new request.
  */
 static void
-http_req_new(size_t id, char *uri, char *modified_since, int count, int outfd)
+http_req_new(unsigned int id, char *uri, char *modified_since, int count,
+    int outfd)
 {
 	struct http_request *req;
 	char *host, *port, *path;
@@ -560,7 +563,7 @@ http_req_free(struct http_request *req)
  * Enqueue request response
  */
 static void
-http_req_done(size_t id, enum http_result res, const char *last_modified)
+http_req_done(unsigned int id, enum http_result res, const char *last_modified)
 {
 	struct ibuf *b;
 
@@ -575,7 +578,7 @@ http_req_done(size_t id, enum http_result res, const char *last_modified)
  * Enqueue request failure response
  */
 static void
-http_req_fail(size_t id)
+http_req_fail(unsigned int id)
 {
 	struct ibuf *b;
 	enum http_result res = HTTP_FAILED;
@@ -1848,8 +1851,11 @@ proc_http(char *bind_addr, int fd)
 				errx(1, "too many connections");
 		}
 
-		if (poll(pfds, i, timeout) == -1)
+		if (poll(pfds, i, timeout) == -1) {
+			if (errno == EINTR)
+				continue;
 			err(1, "poll");
+		}
 
 		if (pfds[0].revents & POLLHUP)
 			break;
@@ -1864,7 +1870,7 @@ proc_http(char *bind_addr, int fd)
 		if (pfds[0].revents & POLLIN) {
 			b = io_buf_recvfd(fd, &inbuf);
 			if (b != NULL) {
-				size_t id;
+				unsigned int id;
 				char *uri;
 				char *mod;
 
