@@ -1,4 +1,4 @@
-/*	$OpenBSD: main.c,v 1.164 2021/11/09 11:03:39 claudio Exp $ */
+/*	$OpenBSD: main.c,v 1.187 2022/01/28 15:30:23 claudio Exp $ */
 /*
  * Copyright (c) 2021 Claudio Jeker <claudio@openbsd.org>
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
@@ -67,7 +67,9 @@ const char	*bird_tablename = "ROAS";
 
 int	verbose;
 int	noop;
+int	filemode;
 int	rrdpon = 1;
+int	repo_timeout;
 
 struct stats	 stats;
 
@@ -103,8 +105,9 @@ entity_free(struct entity *ent)
 	if (ent == NULL)
 		return;
 
-	free(ent->data);
+	free(ent->path);
 	free(ent->file);
+	free(ent->data);
 	free(ent);
 }
 
@@ -117,11 +120,12 @@ void
 entity_read_req(struct ibuf *b, struct entity *ent)
 {
 	io_read_buf(b, &ent->type, sizeof(ent->type));
+	io_read_buf(b, &ent->location, sizeof(ent->location));
+	io_read_buf(b, &ent->repoid, sizeof(ent->repoid));
 	io_read_buf(b, &ent->talid, sizeof(ent->talid));
+	io_read_str(b, &ent->path);
 	io_read_str(b, &ent->file);
-	io_read_buf(b, &ent->has_data, sizeof(ent->has_data));
-	if (ent->has_data)
-		io_read_buf_alloc(b, (void **)&ent->data, &ent->datasz);
+	io_read_buf_alloc(b, (void **)&ent->data, &ent->datasz);
 }
 
 /*
@@ -133,22 +137,43 @@ entity_write_req(const struct entity *ent)
 {
 	struct ibuf *b;
 
-	if (filepath_add(&fpt, ent->file) == 0) {
-		warnx("%s: File already visited", ent->file);
-		entity_queue--;
-		return;
-	}
-
 	b = io_new_buffer();
 	io_simple_buffer(b, &ent->type, sizeof(ent->type));
+	io_simple_buffer(b, &ent->location, sizeof(ent->location));
+	io_simple_buffer(b, &ent->repoid, sizeof(ent->repoid));
 	io_simple_buffer(b, &ent->talid, sizeof(ent->talid));
+	io_str_buffer(b, ent->path);
 	io_str_buffer(b, ent->file);
-	io_simple_buffer(b, &ent->has_data, sizeof(int));
-	if (ent->has_data)
-		io_buf_buffer(b, ent->data, ent->datasz);
+	io_buf_buffer(b, ent->data, ent->datasz);
 	io_close_buffer(&procq, b);
 }
 
+static void
+entity_write_repo(struct repo *rp)
+{
+	struct ibuf *b;
+	enum rtype type = RTYPE_REPO;
+	enum location loc = DIR_UNKNOWN;
+	unsigned int repoid;
+	char *path, *altpath;
+	int talid = 0;
+
+	repoid = repo_id(rp);
+	path = repo_basedir(rp, 0);
+	altpath = repo_basedir(rp, 1);
+	b = io_new_buffer();
+	io_simple_buffer(b, &type, sizeof(type));
+	io_simple_buffer(b, &loc, sizeof(loc));
+	io_simple_buffer(b, &repoid, sizeof(repoid));
+	io_simple_buffer(b, &talid, sizeof(talid));
+	io_str_buffer(b, path);
+	io_str_buffer(b, altpath);
+	io_buf_buffer(b, NULL, 0);
+	io_close_buffer(&procq, b);
+	free(path);
+	free(altpath);
+}
+
 /*
  * Scan through all queued requests and see which ones are in the given
  * repo, then flush those into the parser process.
@@ -158,20 +183,9 @@ entityq_flush(struct entityq *q, struct repo *rp)
 {
 	struct entity	*p, *np;
 
+	entity_write_repo(rp);
+
 	TAILQ_FOREACH_SAFE(p, q, entries, np) {
-		char *file = p->file;
-
-		/*
-		 * XXX fixup path here since the repo may change
-		 * during load because of fallback. In that case
-		 * the file path changes as well since RRDP and RSYNC
-		 * can not share a common repo.
-		 */
-		p->file = repo_filename(rp, file);
-		if (p->file == NULL)
-			err(1, "can't construct repo filename");
-		free(file);
-
 		entity_write_req(p);
 		TAILQ_REMOVE(q, p, entries);
 		entity_free(p);
@@ -182,8 +196,8 @@ entityq_flush(struct entityq *q, struct repo *rp)
  * Add the heap-allocated file to the queue for processing.
  */
 static void
-entityq_add(char *file, enum rtype type, struct repo *rp,
-    unsigned char *data, size_t datasz, int talid)
+entityq_add(char *path, char *file, enum rtype type, enum location loc,
+    struct repo *rp, unsigned char *data, size_t datasz, int talid)
 {
 	struct entity	*p;
 
@@ -191,13 +205,14 @@ entityq_add(char *file, enum rtype type, struct repo *rp,
 		err(1, NULL);
 
 	p->type = type;
+	p->location = loc;
 	p->talid = talid;
+	p->path = path;
+	if (rp != NULL)
+		p->repoid = repo_id(rp);
 	p->file = file;
-	p->has_data = data != NULL;
-	if (p->has_data) {
-		p->data = data;
-		p->datasz = datasz;
-	}
+	p->data = data;
+	p->datasz = (data != NULL) ? datasz : 0;
 
 	entity_queue++;
 
@@ -207,27 +222,13 @@ entityq_add(char *file, enum rtype type, struct repo *rp,
 	 */
 
 	if (rp == NULL || !repo_queued(rp, p)) {
-		/*
-		 * XXX fixup path here since for repo path the
-		 * file path has not yet been fixed here.
-		 * This is a quick way to make this work but in
-		 * the long run repos need to be passed to the parser.
-		 */
-		if (rp != NULL) {
-			file = p->file;
-			p->file = repo_filename(rp, file);
-			if (p->file == NULL)
-				err(1, "can't construct repo filename from %s",
-				    file);
-			free(file);
-		}
 		entity_write_req(p);
 		entity_free(p);
 	}
 }
 
 static void
-rrdp_file_resp(size_t id, int ok)
+rrdp_file_resp(unsigned int id, int ok)
 {
 	enum rrdp_msg type = RRDP_FILE;
 	struct ibuf *b;
@@ -240,7 +241,7 @@ rrdp_file_resp(size_t id, int ok)
 }
 
 void
-rrdp_fetch(size_t id, const char *uri, const char *local,
+rrdp_fetch(unsigned int id, const char *uri, const char *local,
     struct rrdp_session *s)
 {
 	enum rrdp_msg type = RRDP_START;
@@ -261,13 +262,15 @@ rrdp_fetch(size_t id, const char *uri, const char *local,
  * Request a repository sync via rsync URI to directory local.
  */
 void
-rsync_fetch(size_t id, const char *uri, const char *local)
+rsync_fetch(unsigned int id, const char *uri, const char *local,
+    const char *base)
 {
 	struct ibuf	*b;
 
 	b = io_new_buffer();
 	io_simple_buffer(b, &id, sizeof(id));
 	io_str_buffer(b, local);
+	io_str_buffer(b, base);
 	io_str_buffer(b, uri);
 	io_close_buffer(&rsyncq, b);
 }
@@ -276,7 +279,7 @@ rsync_fetch(size_t id, const char *uri, const char *local)
  * Request a file from a https uri, data is written to the file descriptor fd.
  */
 void
-http_fetch(size_t id, const char *uri, const char *last_mod, int fd)
+http_fetch(unsigned int id, const char *uri, const char *last_mod, int fd)
 {
 	struct ibuf	*b;
 
@@ -294,7 +297,7 @@ http_fetch(size_t id, const char *uri, const char *last_mod, int fd)
  * Create a pipe and pass the pipe endpoints to the http and rrdp process.
  */
 static void
-rrdp_http_fetch(size_t id, const char *uri, const char *last_mod)
+rrdp_http_fetch(unsigned int id, const char *uri, const char *last_mod)
 {
 	enum rrdp_msg type = RRDP_HTTP_INI;
 	struct ibuf *b;
@@ -313,7 +316,7 @@ rrdp_http_fetch(size_t id, const char *uri, const char *last_mod)
 }
 
 void
-rrdp_http_done(size_t id, enum http_result res, const char *last_mod)
+rrdp_http_done(unsigned int id, enum http_result res, const char *last_mod)
 {
 	enum rrdp_msg type = RRDP_HTTP_FIN;
 	struct ibuf *b;
@@ -332,23 +335,18 @@ rrdp_http_done(size_t id, enum http_result res, const char *last_mod)
  * These are always relative to the directory in which "mft" sits.
  */
 static void
-queue_add_from_mft(const char *mft, const struct mftfile *file, enum rtype type)
+queue_add_from_mft(const char *path, const struct mftfile *file,
+    struct repo *rp)
 {
-	char		*cp, *nfile;
+	char		*nfile, *npath = NULL;
 
-	/* Construct local path from filename. */
-	cp = strrchr(mft, '/');
-	assert(cp != NULL);
-	assert(cp - mft < INT_MAX);
-	if (asprintf(&nfile, "%.*s/%s", (int)(cp - mft), mft, file->file) == -1)
+	if (path != NULL)
+		if ((npath = strdup(path)) == NULL)
+			err(1, NULL);
+	if ((nfile = strdup(file->file)) == NULL)
 		err(1, NULL);
 
-	/*
-	 * Since we're from the same directory as the MFT file, we know
-	 * that the repository has already been loaded.
-	 */
-
-	entityq_add(nfile, type, NULL, NULL, 0, -1);
+	entityq_add(npath, nfile, file->type, file->location, rp, NULL, 0, -1);
 }
 
 /*
@@ -360,58 +358,54 @@ queue_add_from_mft(const char *mft, const struct mftfile *file, enum rtype type)
  * check the suffix anyway).
  */
 static void
-queue_add_from_mft_set(const struct mft *mft)
+queue_add_from_mft_set(const struct mft *mft, const char *name, struct repo *rp)
 {
-	size_t			 i, sz;
+	size_t			 i;
 	const struct mftfile	*f;
 
 	for (i = 0; i < mft->filesz; i++) {
 		f = &mft->files[i];
-		sz = strlen(f->file);
-		assert(sz > 4);
-		if (strcasecmp(f->file + sz - 4, ".crl") != 0)
+		if (f->type != RTYPE_CRL)
 			continue;
-		queue_add_from_mft(mft->file, f, RTYPE_CRL);
+		queue_add_from_mft(mft->path, f, rp);
 	}
 
 	for (i = 0; i < mft->filesz; i++) {
 		f = &mft->files[i];
-		sz = strlen(f->file);
-		assert(sz > 4);
-		if (strcasecmp(f->file + sz - 4, ".crl") == 0)
+		switch (f->type) {
+		case RTYPE_CER:
+		case RTYPE_ROA:
+		case RTYPE_GBR:
+			queue_add_from_mft(mft->path, f, rp);
+			break;
+		case RTYPE_CRL:
 			continue;
-		else if (strcasecmp(f->file + sz - 4, ".cer") == 0)
-			queue_add_from_mft(mft->file, f, RTYPE_CER);
-		else if (strcasecmp(f->file + sz - 4, ".roa") == 0)
-			queue_add_from_mft(mft->file, f, RTYPE_ROA);
-		else if (strcasecmp(f->file + sz - 4, ".gbr") == 0)
-			queue_add_from_mft(mft->file, f, RTYPE_GBR);
-		else
-			logx("%s: unsupported file type: %s", mft->file,
-			    f->file);
+		default:
+			warnx("%s: unsupported file: %s", name, f->file);
+		}
 	}
 }
 
 /*
- * Add a local TAL file (RFC 7730) to the queue of files to fetch.
+ * Add a local file to the queue of files to fetch.
  */
 static void
-queue_add_tal(const char *file, int id)
+queue_add_file(const char *file, enum rtype type, int talid)
 {
-	unsigned char	*buf;
+	unsigned char	*buf = NULL;
 	char		*nfile;
-	size_t		 len;
+	size_t		 len = 0;
+
+	if (!filemode || strncmp(file, "rsync://", strlen("rsync://")) != 0) {
+		buf = load_file(file, &len);
+		if (buf == NULL)
+			err(1, "%s", file);
+	}
 
 	if ((nfile = strdup(file)) == NULL)
 		err(1, NULL);
-	buf = load_file(file, &len);
-	if (buf == NULL) {
-		warn("%s", file);
-		return;
-	}
-
 	/* Not in a repository, so directly add to queue. */
-	entityq_add(nfile, RTYPE_TAL, NULL, buf, len, id);
+	entityq_add(NULL, nfile, type, DIR_UNKNOWN, NULL, buf, len, talid);
 }
 
 /*
@@ -422,21 +416,30 @@ queue_add_from_tal(struct tal *tal)
 {
 	struct repo	*repo;
 	unsigned char	*data;
+	char		*nfile;
 
 	assert(tal->urisz);
 
 	if ((taldescs[tal->id] = strdup(tal->descr)) == NULL)
 		err(1, NULL);
 
+	/* figure out the TA filename, must be done before repo lookup */
+	nfile = strrchr(tal->uri[0], '/');
+	assert(nfile != NULL);
+	if ((nfile = strdup(nfile + 1)) == NULL)
+		err(1, NULL);
+
 	/* Look up the repository. */
 	repo = ta_lookup(tal->id, tal);
-	if (repo == NULL)
+	if (repo == NULL) {
+		free(nfile);
 		return;
+	}
 
 	/* steal the pkey from the tal structure */
 	data = tal->pkey;
 	tal->pkey = NULL;
-	entityq_add(NULL, RTYPE_CER, repo, data,
+	entityq_add(NULL, nfile, RTYPE_CER, DIR_VALID, repo, data,
 	    tal->pkeysz, tal->id);
 }
 
@@ -447,16 +450,40 @@ static void
 queue_add_from_cert(const struct cert *cert)
 {
 	struct repo	*repo;
-	char		*nfile;
+	char		*nfile, *npath;
+	const char	*uri, *repouri, *file;
+	size_t		 repourisz;
 
 	repo = repo_lookup(cert->talid, cert->repo,
 	    rrdpon ? cert->notify : NULL);
 	if (repo == NULL)
 		return;
 
-	if ((nfile = strdup(cert->mft)) == NULL)
-		err(1, NULL);
-	entityq_add(nfile, RTYPE_MFT, repo, NULL, 0, -1);
+	/*
+	 * Figure out the cert filename and path by chopping up the
+	 * MFT URI in the cert based on the repo base URI.
+	 */
+	uri = cert->mft;
+	repouri = repo_uri(repo);
+	repourisz = strlen(repouri);
+	if (strncmp(repouri, cert->mft, repourisz) != 0) {
+		warnx("%s: URI %s outside of repository", repouri, uri);
+		return;
+	}
+	uri += repourisz + 1;	/* skip base and '/' */
+	file = strrchr(uri, '/');
+	if (file == NULL) {
+		npath = NULL;
+		if ((nfile = strdup(uri)) == NULL)
+			err(1, NULL);
+	} else {
+		if ((npath = strndup(uri, file - uri)) == NULL)
+			err(1, NULL);
+		if ((nfile = strdup(file + 1)) == NULL)
+			err(1, NULL);
+	}
+
+	entityq_add(npath, nfile, RTYPE_MFT, DIR_UNKNOWN, repo, NULL, 0, -1);
 }
 
 /*
@@ -474,6 +501,7 @@ entity_process(struct ibuf *b, struct stats *st, struct vrp_tree *tree,
 	struct cert	*cert;
 	struct mft	*mft;
 	struct roa	*roa;
+	char		*file;
 	int		 c;
 
 	/*
@@ -483,6 +511,16 @@ entity_process(struct ibuf *b, struct stats *st, struct vrp_tree *tree,
 	 * We follow that up with whether the resources didn't parse.
 	 */
 	io_read_buf(b, &type, sizeof(type));
+	io_read_str(b, &file);
+
+	/* in filemode messages can be ignored, only the accounting matters */
+	if (filemode)
+		goto done;
+
+	if (filepath_add(&fpt, file) == 0) {
+		warnx("%s: File already visited", file);
+		goto done;
+	}
 
 	switch (type) {
 	case RTYPE_TAL:
@@ -522,9 +560,11 @@ entity_process(struct ibuf *b, struct stats *st, struct vrp_tree *tree,
 			break;
 		}
 		mft = mft_read(b);
-		if (mft->stale)
+		if (!mft->stale)
+			queue_add_from_mft_set(mft, file,
+			    repo_byid(mft->repoid));
+		else
 			st->mfts_stale++;
-		queue_add_from_mft_set(mft);
 		mft_free(mft);
 		break;
 	case RTYPE_CRL:
@@ -547,10 +587,14 @@ entity_process(struct ibuf *b, struct stats *st, struct vrp_tree *tree,
 	case RTYPE_GBR:
 		st->gbrs++;
 		break;
+	case RTYPE_FILE:
+		break;
 	default:
 		errx(1, "unknown entity type %d", type);
 	}
 
+done:
+	free(file);
 	entity_queue--;
 }
 
@@ -562,7 +606,8 @@ rrdp_process(struct ibuf *b)
 	struct rrdp_session s;
 	char *uri, *last_mod, *data;
 	char hash[SHA256_DIGEST_LENGTH];
-	size_t dsz, id;
+	size_t dsz;
+	unsigned int id;
 	int ok;
 
 	io_read_buf(b, &type, sizeof(type));
@@ -600,6 +645,9 @@ rrdp_process(struct ibuf *b)
 		free(uri);
 		free(data);
 		break;
+	case RRDP_CLEAR:
+		rrdp_clear(id);
+		break;
 	default:
 		errx(1, "unexpected rrdp response");
 	}
@@ -649,7 +697,7 @@ check_fs_size(int fd, const char *cachedir)
 
 	if (fs.f_bavail < minsize / fs.f_frsize || fs.f_favail < minnode) {
 		fprintf(stderr, "WARNING: rpki-client may need more than "
-		    "the availabe disk space\n"
+		    "the available disk space\n"
 		    "on the file-system holding %s.\n", cachedir);
 		fprintf(stderr, "available space: %lldkB, "
 		    "suggested minimum %lldkB\n",
@@ -673,9 +721,9 @@ suicide(int sig __attribute__((unused)))
 int
 main(int argc, char *argv[])
 {
-	int		 rc, c, st, proc, rsync, http, rrdp, ok, hangup = 0;
+	int		 rc, c, st, proc, rsync, http, rrdp, hangup = 0;
 	int		 fl = SOCK_STREAM | SOCK_CLOEXEC | SOCK_NONBLOCK;
-	size_t		 i, id;
+	size_t		 i;
 	pid_t		 pid, procpid, rsyncpid, httppid, rrdppid;
 	int		 fd[2];
 	struct pollfd	 pfd[NPFD];
@@ -686,7 +734,6 @@ main(int argc, char *argv[])
 	char		*bind_addr = NULL;
 	const char	*cachedir = NULL, *outputdir = NULL;
 	const char	*errs, *name;
-	const char	*file = NULL;
 	struct vrp_tree	 vrps = RB_INITIALIZER(&vrps);
 	struct brk_tree  brks = RB_INITIALIZER(&brks);
 	struct rusage	ru;
@@ -708,12 +755,13 @@ main(int argc, char *argv[])
 	}
 	cachedir = RPKI_PATH_BASE_DIR;
 	outputdir = RPKI_PATH_OUT_DIR;
+	repo_timeout = timeout / 4;
 
 	if (pledge("stdio rpath wpath cpath inet fattr dns sendfd recvfd "
 	    "proc exec unveil", NULL) == -1)
 		err(1, "pledge");
 
-	while ((c = getopt(argc, argv, "b:Bcd:e:f:jnorRs:t:T:vV")) != -1)
+	while ((c = getopt(argc, argv, "b:Bcd:e:fjnorRs:t:T:vV")) != -1)
 		switch (c) {
 		case 'b':
 			bind_addr = optarg;
@@ -731,7 +779,7 @@ main(int argc, char *argv[])
 			rsync_prog = optarg;
 			break;
 		case 'f':
-			file = optarg;
+			filemode = 1;
 			noop = 1;
 			break;
 		case 'j':
@@ -753,6 +801,12 @@ main(int argc, char *argv[])
 			timeout = strtonum(optarg, 0, 24*60*60, &errs);
 			if (errs)
 				errx(1, "-s: %s", errs);
+			if (timeout == 0)
+				repo_timeout = 24*60*60;
+			else if (timeout < 1)
+				errx(1, "-s: %i too small", timeout);
+			else
+				repo_timeout = timeout / 4;
 			break;
 		case 't':
 			if (talsz >= TALSZ_MAX)
@@ -776,26 +830,38 @@ main(int argc, char *argv[])
 
 	argv += optind;
 	argc -= optind;
-	if (argc == 1)
-		outputdir = argv[0];
-	else if (argc > 1)
-		goto usage;
 
-	signal(SIGPIPE, SIG_IGN);
+	if (!filemode) {
+		if (argc == 1)
+			outputdir = argv[0];
+		else if (argc > 1)
+			goto usage;
+
+		if (outputdir == NULL) {
+			warnx("output directory required");
+			goto usage;
+		}
+	} else {
+		if (argc == 0)
+			goto usage;
+		outputdir = NULL;
+	}
 
 	if (cachedir == NULL) {
 		warnx("cache directory required");
 		goto usage;
 	}
-	if (outputdir == NULL) {
-		warnx("output directory required");
-		goto usage;
-	}
+
+	signal(SIGPIPE, SIG_IGN);
 
 	if ((cachefd = open(cachedir, O_RDONLY | O_DIRECTORY)) == -1)
 		err(1, "cache directory %s", cachedir);
-	if ((outdirfd = open(outputdir, O_RDONLY | O_DIRECTORY)) == -1)
-		err(1, "output directory %s", outputdir);
+	if (outputdir != NULL) {
+		if ((outdirfd = open(outputdir, O_RDONLY | O_DIRECTORY)) == -1)
+			err(1, "output directory %s", outputdir);
+		if (outformats == 0)
+			outformats = FORMAT_OPENBGPD;
+	}
 
 	check_fs_size(cachefd, cachedir);
 
@@ -1004,7 +1070,12 @@ main(int argc, char *argv[])
 	 */
 
 	for (i = 0; i < talsz; i++)
-		queue_add_tal(tals[i], i);
+		queue_add_file(tals[i], RTYPE_TAL, i);
+
+	if (filemode) {
+		while (*argv != NULL)
+			queue_add_file(*argv++, RTYPE_FILE, 0);
+	}
 
 	/* change working directory to the cache directory */
 	if (fchdir(cachefd) == -1)
@@ -1019,7 +1090,7 @@ main(int argc, char *argv[])
 				pfd[i].events |= POLLOUT;
 		}
 
-		polltim = repo_next_timeout(INFTIM);
+		polltim = repo_check_timeout(INFTIM);
 
 		if ((c = poll(pfd, NPFD, polltim)) == -1) {
 			if (errno == EINTR)
@@ -1051,8 +1122,6 @@ main(int argc, char *argv[])
 		if (hangup)
 			break;
 
-		repo_check_timeout();
-
 		/*
 		 * Check the rsync and http process.
 		 * This means that one of our modules has completed
@@ -1063,6 +1132,9 @@ main(int argc, char *argv[])
 		if ((pfd[1].revents & POLLIN)) {
 			b = io_buf_read(rsync, &rsyncbuf);
 			if (b != NULL) {
+				unsigned int id;
+				int ok;
+
 				io_read_buf(b, &id, sizeof(id));
 				io_read_buf(b, &ok, sizeof(ok));
 				rsync_finish(id, ok);
@@ -1073,6 +1145,7 @@ main(int argc, char *argv[])
 		if ((pfd[2].revents & POLLIN)) {
 			b = io_buf_read(http, &httpbuf);
 			if (b != NULL) {
+				unsigned int id;
 				enum http_result res;
 				char *last_mod;
 
@@ -1163,6 +1236,10 @@ main(int argc, char *argv[])
 	if (entity_queue != 0)
 		errx(1, "not all files processed, giving up");
 
+	/* if processing in filemode the process is done, no cleanup */
+	if (filemode)
+		return rc;
+
 	logx("all files parsed: generating output");
 
 	repo_cleanup(&fpt);
@@ -1202,8 +1279,8 @@ main(int argc, char *argv[])
 	logx("Certificate revocation lists: %zu", stats.crls);
 	logx("Ghostbuster records: %zu", stats.gbrs);
 	logx("Repositories: %zu", stats.repos);
-	logx("Cleanup: removed %zu files, %zu directories",
-	    stats.del_files, stats.del_dirs);
+	logx("Cleanup: removed %zu files, %zu directories, %zu superfluous",
+	    stats.del_files, stats.del_dirs, stats.extra_files);
 	logx("VRP Entries: %zu (%zu unique)", stats.vrps, stats.uniqs);
 
 	/* Memory cleanup. */
@@ -1216,6 +1293,7 @@ usage:
 	    "usage: rpki-client [-BcjnoRrVv] [-b sourceaddr] [-d cachedir]"
 	    " [-e rsync_prog]\n"
 	    "                   [-s timeout] [-T table] [-t tal]"
-	    " [outputdir]\n");
+	    " [outputdir]\n"
+	    "       rpki-client [-Vv] [-d cachedir] [-t tal] -f file ...\n");
 	return 1;
 }
