@@ -1,4 +1,4 @@
-/*	$OpenBSD: mft.c,v 1.42 2021/10/28 13:51:42 job Exp $ */
+/*	$OpenBSD: mft.c,v 1.52 2022/01/28 15:30:23 claudio Exp $ */
 /*
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -16,11 +16,11 @@
  */
 
 #include <assert.h>
+#include <ctype.h>
 #include <err.h>
 #include <limits.h>
 #include <stdarg.h>
 #include <stdint.h>
-#include <fcntl.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
@@ -40,24 +40,7 @@ struct	parse {
 	struct mft	*res; /* result object */
 };
 
-static ASN1_OBJECT    *mft_oid;
-
-static const char *
-gentime2str(const ASN1_GENERALIZEDTIME *time)
-{
-	static char	buf[64];
-	BIO		*mem;
-
-	if ((mem = BIO_new(BIO_s_mem())) == NULL)
-		cryptoerrx("BIO_new");
-	if (!ASN1_GENERALIZEDTIME_print(mem, time))
-		cryptoerrx("ASN1_GENERALIZEDTIME_print");
-	if (BIO_gets(mem, buf, sizeof(buf)) < 0)
-		cryptoerrx("BIO_gets");
-
-	BIO_free(mem);
-	return buf;
-}
+extern ASN1_OBJECT    *mft_oid;
 
 /*
  * Convert an ASN1_GENERALIZEDTIME to a struct tm.
@@ -79,48 +62,107 @@ generalizedtime_to_tm(const ASN1_GENERALIZEDTIME *gtime, struct tm *tm)
 
 /*
  * Validate and verify the time validity of the mft.
- * Returns 1 if all is good, 0 if mft is stale, any other case -1.
+ * Returns 1 if all is good and for any other case 0.
  */
 static int
-check_validity(const ASN1_GENERALIZEDTIME *from,
-    const ASN1_GENERALIZEDTIME *until, const char *fn)
+mft_parse_time(const ASN1_GENERALIZEDTIME *from,
+    const ASN1_GENERALIZEDTIME *until, struct parse *p)
 {
-	time_t now = time(NULL);
-	struct tm tm_from, tm_until, tm_now;
-
-	if (gmtime_r(&now, &tm_now) == NULL) {
-		warnx("%s: could not get current time", fn);
-		return -1;
-	}
+	struct tm tm_from, tm_until;
 
 	if (!generalizedtime_to_tm(from, &tm_from)) {
-		warnx("%s: embedded from time format invalid", fn);
-		return -1;
+		warnx("%s: embedded from time format invalid", p->fn);
+		return 0;
 	}
 	if (!generalizedtime_to_tm(until, &tm_until)) {
-		warnx("%s: embedded until time format invalid", fn);
-		return -1;
+		warnx("%s: embedded until time format invalid", p->fn);
+		return 0;
 	}
 
 	/* check that until is not before from */
 	if (ASN1_time_tm_cmp(&tm_until, &tm_from) < 0) {
-		warnx("%s: bad update interval", fn);
-		return -1;
-	}
-	/* check that now is not before from */
-	if (ASN1_time_tm_cmp(&tm_from, &tm_now) > 0) {
-		warnx("%s: mft not yet valid %s", fn, gentime2str(from));
-		return -1;
-	}
-	/* check that now is not after until */
-	if (ASN1_time_tm_cmp(&tm_until, &tm_now) < 0) {
-		warnx("%s: mft expired on %s", fn, gentime2str(until));
+		warnx("%s: bad update interval", p->fn);
 		return 0;
 	}
 
+	if ((p->res->valid_from = mktime(&tm_from)) == -1 ||
+	    (p->res->valid_until = mktime(&tm_until)) == -1)
+		errx(1, "%s: mktime failed", p->fn);
+
 	return 1;
 }
 
+/*
+ * Determine rtype corresponding to file extension. Returns RTYPE_INVALID
+ * on error or unkown extension.
+ */
+enum rtype
+rtype_from_file_extension(const char *fn)
+{
+	size_t	 sz;
+
+	sz = strlen(fn);
+	if (sz < 5)
+		return RTYPE_INVALID;
+
+	if (strcasecmp(fn + sz - 4, ".tal") == 0)
+		return RTYPE_TAL;
+	if (strcasecmp(fn + sz - 4, ".cer") == 0)
+		return RTYPE_CER;
+	if (strcasecmp(fn + sz - 4, ".crl") == 0)
+		return RTYPE_CRL;
+	if (strcasecmp(fn + sz - 4, ".mft") == 0)
+		return RTYPE_MFT;
+	if (strcasecmp(fn + sz - 4, ".roa") == 0)
+		return RTYPE_ROA;
+	if (strcasecmp(fn + sz - 4, ".gbr") == 0)
+		return RTYPE_GBR;
+
+	return RTYPE_INVALID;
+}
+
+/*
+ * Validate that a filename listed on a Manifest only contains characters
+ * permitted in draft-ietf-sidrops-6486bis section 4.2.2
+ */
+static int
+valid_filename(const char *fn, size_t len)
+{
+	const unsigned char *c;
+	size_t i;
+
+	for (c = fn, i = 0; i < len; i++, c++)
+		if (!isalnum(*c) && *c != '-' && *c != '_' && *c != '.')
+			return 0;
+
+	c = memchr(fn, '.', len);
+	if (c == NULL || c != memrchr(fn, '.', len))
+		return 0;
+
+	return 1;
+}
+
+/*
+ * Check that the file is a CER, CRL, GBR or a ROA.
+ * Returns corresponding rtype or RTYPE_INVALID on error.
+ */
+static enum rtype
+rtype_from_mftfile(const char *fn)
+{
+	enum rtype		 type;
+
+	type = rtype_from_file_extension(fn);
+	switch (type) {
+	case RTYPE_CER:
+	case RTYPE_CRL:
+	case RTYPE_GBR:
+	case RTYPE_ROA:
+		return type;
+	default:
+		return RTYPE_INVALID;
+	}
+}
+
 /*
  * Parse an individual "FileAndHash", RFC 6486, sec. 4.2.
  * Return zero on failure, non-zero on success.
@@ -156,28 +198,16 @@ mft_parse_filehash(struct parse *p, const ASN1_OCTET_STRING *os)
 		    p->fn, ASN1_tag2str(file->type), file->type);
 		goto out;
 	}
+	if (!valid_filename(file->value.ia5string->data,
+	    file->value.ia5string->length)) {
+		warnx("%s: RFC 6486 section 4.2.2: bad filename", p->fn);
+		goto out;
+	}
 	fn = strndup((const char *)file->value.ia5string->data,
 	    file->value.ia5string->length);
 	if (fn == NULL)
 		err(1, NULL);
 
-	/*
-	 * Make sure we're just a pathname and either an ROA or CER.
-	 * I don't think that the RFC specifically mentions this, but
-	 * it's in practical use and would really screw things up
-	 * (arbitrary filenames) otherwise.
-	 */
-
-	if (strchr(fn, '/') != NULL) {
-		warnx("%s: path components disallowed in filename: %s",
-		    p->fn, fn);
-		goto out;
-	} else if (strlen(fn) <= 4) {
-		warnx("%s: filename must be large enough for suffix part: %s",
-		    p->fn, fn);
-		goto out;
-	}
-
 	/* Now hash value. */
 
 	hash = sk_ASN1_TYPE_value(seq, 1);
@@ -198,6 +228,7 @@ mft_parse_filehash(struct parse *p, const ASN1_OCTET_STRING *os)
 	/* Insert the filename and hash value. */
 	fent = &p->res->files[p->res->filesz++];
 
+	fent->type = rtype_from_mftfile(fn);
 	fent->file = fn;
 	fn = NULL;
 	memcpy(fent->hash, hash->value.bit_string->data, SHA256_DIGEST_LENGTH);
@@ -257,7 +288,7 @@ mft_parse_flist(struct parse *p, const ASN1_OCTET_STRING *os)
 
 /*
  * Handle the eContent of the manifest object, RFC 6486 sec. 4.2.
- * Returns <0 on failure, 0 on stale, >0 on success.
+ * Returns 0 on failure and 1 on success.
  */
 static int
 mft_parse_econtent(const unsigned char *d, size_t dsz, struct parse *p)
@@ -267,7 +298,7 @@ mft_parse_econtent(const unsigned char *d, size_t dsz, struct parse *p)
 	const ASN1_GENERALIZEDTIME *from, *until;
 	long			 mft_version;
 	BIGNUM			*mft_seqnum = NULL;
-	int			 i = 0, rc = -1;
+	int			 i = 0, rc = 0;
 
 	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
 		cryptowarnx("%s: RFC 6486 section 4.2: Manifest: "
@@ -366,22 +397,19 @@ mft_parse_econtent(const unsigned char *d, size_t dsz, struct parse *p)
 	}
 	until = t->value.generalizedtime;
 
-	rc = check_validity(from, until, p->fn);
-	if (rc != 1)
+	if (!mft_parse_time(from, until, p))
 		goto out;
 
-	/* The mft is valid. Reset rc so later 'goto out' return failure. */
-	rc = -1;
-
 	/* File list algorithm. */
 
 	t = sk_ASN1_TYPE_value(seq, i++);
 	if (t->type != V_ASN1_OBJECT) {
 		warnx("%s: RFC 6486 section 4.2.1: fileHashAlg: "
-		    "want ASN.1 object time, have %s (NID %d)",
+		    "want ASN.1 object, have %s (NID %d)",
 		    p->fn, ASN1_tag2str(t->type), t->type);
 		goto out;
-	} else if (OBJ_obj2nid(t->value.object) != NID_sha256) {
+	}
+	if (OBJ_obj2nid(t->value.object) != NID_sha256) {
 		warnx("%s: RFC 6486 section 4.2.1: fileHashAlg: "
 		    "want SHA256 object, have %s (NID %d)", p->fn,
 		    ASN1_tag2str(OBJ_obj2nid(t->value.object)),
@@ -397,7 +425,9 @@ mft_parse_econtent(const unsigned char *d, size_t dsz, struct parse *p)
 		    "want ASN.1 sequence, have %s (NID %d)",
 		    p->fn, ASN1_tag2str(t->type), t->type);
 		goto out;
-	} else if (!mft_parse_flist(p, t->value.octet_string))
+	}
+
+	if (!mft_parse_flist(p, t->value.octet_string))
 		goto out;
 
 	rc = 1;
@@ -418,20 +448,13 @@ struct mft *
 mft_parse(X509 **x509, const char *fn, const unsigned char *der, size_t len)
 {
 	struct parse	 p;
-	int		 c, rc = 0;
-	size_t		 i, cmsz;
+	int		 rc = 0;
+	size_t		 cmsz;
 	unsigned char	*cms;
 
 	memset(&p, 0, sizeof(struct parse));
 	p.fn = fn;
 
-	if (mft_oid == NULL) {
-		mft_oid = OBJ_txt2obj("1.2.840.113549.1.9.16.1.26", 1);
-		if (mft_oid == NULL)
-			errx(1, "OBJ_txt2obj for %s failed",
-			    "1.2.840.113549.1.9.16.1.26");
-	}
-
 	cms = cms_parse_validate(x509, fn, der, len, mft_oid, &cmsz);
 	if (cms == NULL)
 		return NULL;
@@ -439,8 +462,6 @@ mft_parse(X509 **x509, const char *fn, const unsigned char *der, size_t len)
 
 	if ((p.res = calloc(1, sizeof(struct mft))) == NULL)
 		err(1, NULL);
-	if ((p.res->file = strdup(fn)) == NULL)
-		err(1, NULL);
 
 	p.res->aia = x509_get_aia(*x509, fn);
 	p.res->aki = x509_get_aki(*x509, 0, fn);
@@ -451,27 +472,7 @@ mft_parse(X509 **x509, const char *fn, const unsigned char *der, size_t len)
 		goto out;
 	}
 
-	/*
-	 * If we're stale, then remove all of the files that the MFT
-	 * references as well as marking it as stale.
-	 */
-
-	if ((c = mft_parse_econtent(cms, cmsz, &p)) == 0) {
-		/*
-		 * FIXME: it should suffice to just mark this as stale
-		 * and have the logic around mft_read() simply ignore
-		 * the contents of stale entries, just like it does for
-		 * invalid ROAs or certificates.
-		 */
-
-		p.res->stale = 1;
-		if (p.res->files != NULL)
-			for (i = 0; i < p.res->filesz; i++)
-				free(p.res->files[i].file);
-		free(p.res->files);
-		p.res->filesz = 0;
-		p.res->files = NULL;
-	} else if (c == -1)
+	if (mft_parse_econtent(cms, cmsz, &p) == 0)
 		goto out;
 
 	rc = 1;
@@ -486,44 +487,6 @@ out:
 	return p.res;
 }
 
-/*
- * Check all files and their hashes in a MFT structure.
- * Return zero on failure, non-zero on success.
- */
-int
-mft_check(const char *fn, struct mft *p)
-{
-	size_t	i;
-	int	rc = 1;
-	char	*cp, *h, *path = NULL;
-
-	/* Check hash of file now, but first build path for it */
-	cp = strrchr(fn, '/');
-	assert(cp != NULL);
-	assert(cp - fn < INT_MAX);
-
-	for (i = 0; i < p->filesz; i++) {
-		const struct mftfile *m = &p->files[i];
-		if (!valid_filename(m->file)) {
-			if (base64_encode(m->hash, sizeof(m->hash), &h) == -1)
-				errx(1, "base64_encode failed in %s", __func__);
-			warnx("%s: unsupported filename for %s", fn, h);
-			free(h);
-			continue;
-		}
-		if (asprintf(&path, "%.*s/%s", (int)(cp - fn), fn,
-		    m->file) == -1)
-			err(1, NULL);
-		if (!valid_filehash(path, m->hash, sizeof(m->hash))) {
-			warnx("%s: bad message digest for %s", fn, m->file);
-			rc = 0;
-		}
-		free(path);
-	}
-
-	return rc;
-}
-
 /*
  * Free an MFT pointer.
  * Safe to call with NULL.
@@ -543,7 +506,7 @@ mft_free(struct mft *p)
 	free(p->aia);
 	free(p->aki);
 	free(p->ski);
-	free(p->file);
+	free(p->path);
 	free(p->files);
 	free(p->seqnum);
 	free(p);
@@ -558,18 +521,23 @@ mft_buffer(struct ibuf *b, const struct mft *p)
 {
 	size_t		 i;
 
-	io_simple_buffer(b, &p->stale, sizeof(int));
-	io_str_buffer(b, p->file);
-	io_simple_buffer(b, &p->filesz, sizeof(size_t));
-
-	for (i = 0; i < p->filesz; i++) {
-		io_str_buffer(b, p->files[i].file);
-		io_simple_buffer(b, p->files[i].hash, SHA256_DIGEST_LENGTH);
-	}
+	io_simple_buffer(b, &p->stale, sizeof(p->stale));
+	io_simple_buffer(b, &p->repoid, sizeof(p->repoid));
+	io_str_buffer(b, p->path);
 
 	io_str_buffer(b, p->aia);
 	io_str_buffer(b, p->aki);
 	io_str_buffer(b, p->ski);
+
+	io_simple_buffer(b, &p->filesz, sizeof(size_t));
+	for (i = 0; i < p->filesz; i++) {
+		io_str_buffer(b, p->files[i].file);
+		io_simple_buffer(b, &p->files[i].type,
+		    sizeof(p->files[i].type));
+		io_simple_buffer(b, &p->files[i].location,
+		    sizeof(p->files[i].location));
+		io_simple_buffer(b, p->files[i].hash, SHA256_DIGEST_LENGTH);
+	}
 }
 
 /*
@@ -585,23 +553,52 @@ mft_read(struct ibuf *b)
 	if ((p = calloc(1, sizeof(struct mft))) == NULL)
 		err(1, NULL);
 
-	io_read_buf(b, &p->stale, sizeof(int));
-	io_read_str(b, &p->file);
-	io_read_buf(b, &p->filesz, sizeof(size_t));
-
-	assert(p->file);
-	if ((p->files = calloc(p->filesz, sizeof(struct mftfile))) == NULL)
-		err(1, NULL);
-
-	for (i = 0; i < p->filesz; i++) {
-		io_read_str(b, &p->files[i].file);
-		io_read_buf(b, p->files[i].hash, SHA256_DIGEST_LENGTH);
-	}
+	io_read_buf(b, &p->stale, sizeof(p->stale));
+	io_read_buf(b, &p->repoid, sizeof(p->repoid));
+	io_read_str(b, &p->path);
 
 	io_read_str(b, &p->aia);
 	io_read_str(b, &p->aki);
 	io_read_str(b, &p->ski);
 	assert(p->aia && p->aki && p->ski);
 
+	io_read_buf(b, &p->filesz, sizeof(size_t));
+	if ((p->files = calloc(p->filesz, sizeof(struct mftfile))) == NULL)
+		err(1, NULL);
+
+	for (i = 0; i < p->filesz; i++) {
+		io_read_str(b, &p->files[i].file);
+		io_read_buf(b, &p->files[i].type, sizeof(p->files[i].type));
+		io_read_buf(b, &p->files[i].location,
+		    sizeof(p->files[i].location));
+		io_read_buf(b, p->files[i].hash, SHA256_DIGEST_LENGTH);
+	}
+
 	return p;
 }
+
+/*
+ * Compare two MFT files, returns 1 if first MFT is preferred and 0 if second
+ * MFT should be used.
+ */
+int
+mft_compare(const struct mft *a, const struct mft *b)
+{
+	int r;
+
+	if (b == NULL)
+		return 1;
+	if (a == NULL)
+		return 0;
+
+	r = strlen(a->seqnum) - strlen(b->seqnum);
+	if (r > 0)	/* seqnum in a is longer -> higher */
+		return 1;
+	if (r < 0)	/* seqnum in a is shorter -> smaller */
+		return 0;
+
+	r = strcmp(a->seqnum, b->seqnum);
+	if (r >= 0)	/* a is greater or equal, prefer a */
+		return 1;
+	return 0;
+}
