@@ -1,4 +1,4 @@
-/*	$OpenBSD: roa.c,v 1.32 2021/11/05 10:50:41 claudio Exp $ */
+/*	$OpenBSD: roa.c,v 1.37 2022/01/18 16:29:06 claudio Exp $ */
 /*
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -36,7 +36,7 @@ struct	parse {
 	struct roa	 *res; /* results */
 };
 
-static ASN1_OBJECT	*roa_oid;
+extern ASN1_OBJECT	*roa_oid;
 
 /*
  * Parse IP address (ROAIPAddress), RFC 6482, section 3.3.
@@ -50,7 +50,7 @@ roa_parse_addr(const ASN1_OCTET_STRING *os, enum afi afi, struct parse *p)
 	size_t			 dsz = os->length;
 	int			 rc = 0;
 	const ASN1_TYPE		*t;
-	const ASN1_INTEGER	*maxlength = NULL;
+	const ASN1_INTEGER	*maxlength;
 	long			 maxlen;
 	struct ip_addr		 addr;
 	struct roa_ip		*res;
@@ -83,6 +83,7 @@ roa_parse_addr(const ASN1_OCTET_STRING *os, enum afi afi, struct parse *p)
 		    "invalid IP address", p->fn);
 		goto out;
 	}
+	maxlen = addr.prefixlen;
 
 	if (sk_ASN1_TYPE_num(seq) == 2) {
 		t = sk_ASN1_TYPE_value(seq, 1);
@@ -115,7 +116,7 @@ roa_parse_addr(const ASN1_OCTET_STRING *os, enum afi afi, struct parse *p)
 
 	res->addr = addr;
 	res->afi = afi;
-	res->maxlength = (maxlength == NULL) ? addr.prefixlen : maxlen;
+	res->maxlength = maxlen;
 	ip_roa_compose_ranges(res);
 
 	rc = 1;
@@ -180,6 +181,11 @@ roa_parse_ipfam(const ASN1_OCTET_STRING *os, struct parse *p)
 	}
 
 	/* will be called multiple times so use recallocarray */
+	if (p->res->ipsz + sk_ASN1_TYPE_num(sseq) >= MAX_IP_SIZE) {
+		warnx("%s: too many IPAddress entries: limit %d",
+		    p->fn, MAX_IP_SIZE);
+		goto out;
+	}
 	p->res->ips = recallocarray(p->res->ips, p->res->ipsz,
 	    p->res->ipsz + sk_ASN1_TYPE_num(sseq), sizeof(struct roa_ip));
 	if (p->res->ips == NULL)
@@ -340,14 +346,6 @@ roa_parse(X509 **x509, const char *fn, const unsigned char *der, size_t len)
 	memset(&p, 0, sizeof(struct parse));
 	p.fn = fn;
 
-	/* OID from section 2, RFC 6482. */
-	if (roa_oid == NULL) {
-		roa_oid = OBJ_txt2obj("1.2.840.113549.1.9.16.1.24", 1);
-		if (roa_oid == NULL)
-			errx(1, "OBJ_txt2obj for %s failed",
-			    "1.2.840.113549.1.9.16.1.24");
-	}
-
 	cms = cms_parse_validate(x509, fn, der, len, roa_oid, &cmsz);
 	if (cms == NULL)
 		return NULL;
