@@ -1,4 +1,4 @@
-/*	$OpenBSD: rrdp_notification.c,v 1.11 2021/11/09 11:01:04 claudio Exp $ */
+/*	$OpenBSD: rrdp_notification.c,v 1.13 2022/02/03 18:19:32 claudio Exp $ */
 /*
  * Copyright (c) 2020 Nils Fisher <nils_fisher@hotmail.com>
  * Copyright (c) 2021 Claudio Jeker <claudio@openbsd.org>
@@ -120,7 +120,8 @@ start_notification_elem(struct notification_xml *nxml, const char **attr)
 		    "parse failed - entered notification elem unexpectedely");
 	for (i = 0; attr[i]; i += 2) {
 		const char *errstr;
-		if (strcmp("xmlns", attr[i]) == 0) {
+		if (strcmp("xmlns", attr[i]) == 0 &&
+		    strcmp(RRDP_XMLNS, attr[i + 1]) == 0) {
 			has_xmlns = 1;
 			continue;
 		}
@@ -464,6 +465,19 @@ notification_delta_done(struct notification_xml *nxml)
 void
 log_notification_xml(struct notification_xml *nxml)
 {
+	struct delta_item *d;
+	char *hash;
+
 	logx("session_id: %s, serial: %lld", nxml->session_id, nxml->serial);
 	logx("snapshot_uri: %s", nxml->snapshot_uri);
+	hash = hex_encode(nxml->snapshot_hash, sizeof(nxml->snapshot_hash));
+	logx("snapshot hash: %s", hash);
+	free(hash);
+
+	TAILQ_FOREACH(d, &nxml->delta_q, q) {
+		logx("delta serial %lld uri: %s", d->serial, d->uri);
+		hash = hex_encode(d->hash, sizeof(d->hash));
+		logx("delta hash: %s", hash);
+		free(hash);
+	}
 }
