@@ -1,4 +1,4 @@
-/*	$OpenBSD: tal.c,v 1.32 2021/10/26 16:12:54 claudio Exp $ */
+/*	$OpenBSD: tal.c,v 1.34 2021/11/04 11:32:55 claudio Exp $ */
 /*
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -58,15 +58,17 @@ tal_parse_buffer(const char *fn, char *buf, size_t len)
 	while ((nl = memchr(buf, '\n', len)) != NULL) {
 		line = buf;
 
-		/* replace LF and optional CR with NUL */
-		*nl = '\0';
-		if (nl > line && nl[-1] == '\r')
-			nl[-1] = '\0';
-
 		/* advance buffer to next line */
 		len -= nl + 1 - buf;
 		buf = nl + 1;
 
+		/* replace LF and optional CR with NUL, point nl at first NUL */
+		*nl = '\0';
+		if (nl > line && nl[-1] == '\r') {
+			nl[-1] = '\0';
+			nl--;
+		}
+
 		if (optcomment) {
 			/* if this is a comment, just eat the line */
 			if (line[0] == '#')
@@ -213,9 +215,10 @@ tal_buffer(struct ibuf *b, const struct tal *p)
 {
 	size_t	 i;
 
+	io_simple_buffer(b, &p->id, sizeof(p->id));
 	io_buf_buffer(b, p->pkey, p->pkeysz);
 	io_str_buffer(b, p->descr);
-	io_simple_buffer(b, &p->urisz, sizeof(size_t));
+	io_simple_buffer(b, &p->urisz, sizeof(p->urisz));
 
 	for (i = 0; i < p->urisz; i++)
 		io_str_buffer(b, p->uri[i]);
@@ -235,9 +238,10 @@ tal_read(struct ibuf *b)
 	if ((p = calloc(1, sizeof(struct tal))) == NULL)
 		err(1, NULL);
 
+	io_read_buf(b, &p->id, sizeof(p->id));
 	io_read_buf_alloc(b, (void **)&p->pkey, &p->pkeysz);
 	io_read_str(b, &p->descr);
-	io_read_buf(b, &p->urisz, sizeof(size_t));
+	io_read_buf(b, &p->urisz, sizeof(p->urisz));
 	assert(p->pkeysz > 0);
 	assert(p->descr);
 	assert(p->urisz > 0);
