@@ -1,4 +1,4 @@
-/*	$OpenBSD: x509.c,v 1.29 2021/10/28 09:02:19 beck Exp $ */
+/*	$OpenBSD: x509.c,v 1.34 2022/02/04 16:08:53 tb Exp $ */
 /*
  * Copyright (c) 2021 Claudio Jeker <claudio@openbsd.org>
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
@@ -30,11 +30,36 @@
 
 #include "extern.h"
 
-static ASN1_OBJECT	*bgpsec_oid;	/* id-kp-bgpsec-router */
+ASN1_OBJECT	*certpol_oid;	/* id-cp-ipAddr-asNumber cert policy */
+ASN1_OBJECT	*carepo_oid;	/* 1.3.6.1.5.5.7.48.5 (caRepository) */
+ASN1_OBJECT	*manifest_oid;	/* 1.3.6.1.5.5.7.48.10 (rpkiManifest) */
+ASN1_OBJECT	*notify_oid;	/* 1.3.6.1.5.5.7.48.13 (rpkiNotify) */
+ASN1_OBJECT	*roa_oid;	/* id-ct-routeOriginAuthz CMS content type */
+ASN1_OBJECT	*mft_oid;	/* id-ct-rpkiManifest CMS content type */
+ASN1_OBJECT	*gbr_oid;	/* id-ct-rpkiGhostbusters CMS content type */
+ASN1_OBJECT	*bgpsec_oid;	/* id-kp-bgpsec-router Key Purpose */
 
-static void
-init_oid(void)
+void
+x509_init_oid(void)
 {
+
+	if ((certpol_oid = OBJ_txt2obj("1.3.6.1.5.5.7.14.2", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.14.2");
+	if ((carepo_oid = OBJ_txt2obj("1.3.6.1.5.5.7.48.5", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.48.5");
+	if ((manifest_oid = OBJ_txt2obj("1.3.6.1.5.5.7.48.10", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.48.10");
+	if ((notify_oid = OBJ_txt2obj("1.3.6.1.5.5.7.48.13", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.48.13");
+	if ((roa_oid = OBJ_txt2obj("1.2.840.113549.1.9.16.1.24", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed",
+		    "1.2.840.113549.1.9.16.1.24");
+	if ((mft_oid = OBJ_txt2obj("1.2.840.113549.1.9.16.1.26", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed",
+		    "1.2.840.113549.1.9.16.1.26");
+	if ((gbr_oid = OBJ_txt2obj("1.2.840.113549.1.9.16.1.35", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed",
+		    "1.2.840.113549.1.9.16.1.35");
 	if ((bgpsec_oid = OBJ_txt2obj("1.3.6.1.5.5.7.3.30", 1)) == NULL)
 		errx(1, "OBJ_txt2obj for %s failed", "1.3.6.1.5.5.7.3.30");
 }
@@ -167,9 +192,6 @@ x509_get_purpose(X509 *x, const char *fn)
 		goto out;
 	}
 
-	if (bgpsec_oid == NULL)
-		init_oid();
-
 	if (OBJ_cmp(bgpsec_oid, sk_ASN1_OBJECT_value(eku, 0)) == 0) {
 		purpose = CERT_PURPOSE_BGPSEC_ROUTER;
 		goto out;
@@ -197,7 +219,7 @@ x509_get_pubkey(X509 *x, const char *fn)
 
 	pkey = X509_get0_pubkey(x);
 	if (pkey == NULL) {
-		warnx("%s: X509_get_pubkey failed in %s", fn, __func__);
+		warnx("%s: X509_get0_pubkey failed in %s", fn, __func__);
 		goto out;
 	}
 	if (EVP_PKEY_base_id(pkey) != EVP_PKEY_EC) {
@@ -331,7 +353,7 @@ x509_get_expire(X509 *x, const char *fn, time_t *tt)
 /*
  * Parse the very specific subset of information in the CRL distribution
  * point extension.
- * See RFC 6487, sectoin 4.8.6 for details.
+ * See RFC 6487, section 4.8.6 for details.
  * Returns NULL on failure, the crl URI on success which has to be freed
  * after use.
  */
